@@ -21,29 +21,32 @@
 **  only under the terms of either the Artistic License or the GNU General
 **  Public License, which may be found in the ePerl source distribution.
 **  Look at the files ARTISTIC and COPYING or run ``eperl -l'' to receive
-**  a builtin copy of both license files.
+**  a built-in copy of both license files.
 **
 **  This program is distributed in the hope that it will be useful, but
 **  WITHOUT ANY WARRANTY; without even the implied warranty of
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See either the
-**  GNU General Public License or the Artistic License for more details.
+**  Artistic License or the GNU General Public License for more details.
 **
 **  ======================================================================
 **
-**  eperl_nphcgi.c -- ePerl stand-alone NPH-CGI/1.1 program
+**  eperl_main.c -- ePerl main procedure 
 */
 
 #include "eperl_global.h"
 #include "eperl_security.h"
 #include "eperl_proto.h"
+#include "eperl_getopt.h"
 
 #include <EXTERN.h>
 #include <perl.h>                 
 
-int mode = MODE_UNKNOWN;
+/* Perl/SFIO kludge */
+#ifdef USE_SFIO
+#define fwrite PerlIO_write
+#endif
 
-char *begin_delimiter = NULL;
-char   *end_delimiter = NULL;
+int mode = MODE_UNKNOWN;
 
 char *allowed_file_ext[]   = LIST_OF_ALLOWED_FILE_EXT;
 char *allowed_caller_uid[] = LIST_OF_ALLOWED_CALLER_UID;
@@ -66,7 +69,7 @@ void PrintError(int mode, char *scripturl, char *scriptfile, char *logfile, char
 
     if (mode == MODE_CGI || mode == MODE_NPHCGI) {
         if (mode == MODE_NPHCGI)
-            PrintHTTPResponse();
+            HTTP_PrintResponseHeaders();
         printf("Content-Type: text/html\n\n");
         printf("<html>\n");
         printf("<head>\n");
@@ -77,6 +80,7 @@ void PrintError(int mode, char *scripturl, char *scriptfile, char *logfile, char
         cp = getenv("SCRIPT_NAME");
         if (cp == NULL)
             cp = "UNKNOWN_IMG_DIR";
+        printf("<a href=\"http://www.engelschall.com/sw/eperl/\"><img src=\"%s/powered.gif\" alt=\"Powered By ePerl\" align=right border=0></a>\n", cp);
         printf("<table cellspacing=0 cellpadding=0 border=0>\n");
         printf("<tr>\n");
         printf("<td><img src=\"%s/logo.gif\" alt=\"Embedded Perl 5 Language\"></td>\n", cp);
@@ -86,12 +90,12 @@ void PrintError(int mode, char *scripturl, char *scriptfile, char *logfile, char
         printf("</tr>\n");
         printf("</table>\n");
         printf("<p>\n");
-        printf("<table bgcolor=\"#f0d0d0\" cellspacing=0 cellpadding=10 border=0>\n");
-        printf("<tr><td bgcolor=\"#d0c0c0\">\n");
+        printf("<table bgcolor=\"#d0d0f0\" cellspacing=0 cellpadding=10 border=0>\n");
+        printf("<tr><td bgcolor=\"#b0b0d0\">\n");
         printf("<font face=\"Arial, Helvetica\"><b>ERROR:</b></font>\n");
         printf("</td></tr>\n");
         printf("<tr><td>\n");
-        printf("<h1><font color=\"#cc3333\">%s</font></h1>\n", ca);
+        printf("<h1><font color=\"#3333cc\">%s</font></h1>\n", ca);
         printf("</td></tr>\n");
         printf("</table>\n");
         if (logfile != NULL) {
@@ -142,7 +146,7 @@ void give_version(void)
     fprintf(stdout, "This program is distributed in the hope that it will be useful,\n");
     fprintf(stdout, "but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
     fprintf(stdout, "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See either\n");
-    fprintf(stdout, "the GNU General Public License or the Artistic License for more details.\n");
+    fprintf(stdout, "the Artistic License or the GNU General Public License for more details.\n");
     fprintf(stdout, "\n");
 }
 
@@ -151,10 +155,10 @@ void give_version_extended(void)
     fprintf(stdout, "%s\n", ePerl_Hello);
     fprintf(stdout, "\n");
     fprintf(stdout, "Characteristics of this binary:\n");
-    fprintf(stdout, "  Perl 5 Version    : %s (%s)\n", AC_perlvers, AC_perlprog);
-    fprintf(stdout, "  Perl 5 Library    : %s/CORE/libperl.a\n", AC_perlarch);
-    fprintf(stdout, "  Perl 5 DynaLoader : %s\n", AC_perldla);
-    fprintf(stdout, "  Additional Libs   : %s\n", AC_perllibs);
+    fprintf(stdout, "  Perl 5 Version    : %s (%s)\n", AC_perl_vers, AC_perl_prog);
+    fprintf(stdout, "  Perl 5 Library    : %s/CORE/libperl.a\n", AC_perl_archlib);
+    fprintf(stdout, "  Perl 5 DynaLoader : %s\n", AC_perl_dla);
+    fprintf(stdout, "  Additional Libs   : %s\n", AC_perl_libs);
     fprintf(stdout, "\n");
 }
 
@@ -170,69 +174,45 @@ void give_license(void)
 
 void give_img_logo(void)
 {
-    if (mode == MODE_NPHCGI) {
-        PrintHTTPResponse();
-    }
+    if (mode == MODE_NPHCGI)
+        HTTP_PrintResponseHeaders();
     printf("Content-Type: image/gif\n\n");
     fwrite(ePerl_LOGO_data, ePerl_LOGO_size, 1, stdout);
 }
 
+void give_img_powered(void)
+{
+    if (mode == MODE_NPHCGI)
+        HTTP_PrintResponseHeaders();
+    printf("Content-Type: image/gif\n\n");
+    fwrite(ePerl_POWERED_data, ePerl_POWERED_size, 1, stdout);
+}
+
 void give_usage(char *name)
 {
-    give_version();
     fprintf(stderr, "Usage: %s [options] [file]\n", name);
     fprintf(stderr, "  where options are:\n");
     fprintf(stderr, "   -d name=value  define global Perl variable ($main::name)\n");
     fprintf(stderr, "   -D name=value  define environment variable ($ENV{'name'})\n");
     fprintf(stderr, "   -B str         set begin block delimiter\n");
     fprintf(stderr, "   -E str         set end block delimiter\n");
+    fprintf(stderr, "   -i             block delimiters are case-insensitive\n");
     fprintf(stderr, "   -m f|c|n       force runtime mode to FILTER, CGI or NPH-CGI\n");
     fprintf(stderr, "   -o outputfile  force the output to be send to this file (default=stdout)\n");
     fprintf(stderr, "   -k             force keeping of current working directory\n");
     fprintf(stderr, "   -x             force debugging output to console (/dev/tty)\n");
+    fprintf(stderr, "   -I directory   specify @INC/#include directory\n");
+    fprintf(stderr, "   -P             enable ePerl Preprocessor\n");
+    fprintf(stderr, "   -C             enable HTML entity conversion for ePerl blocks\n");
+    fprintf(stderr, "   -L             enable line continuation via backslashes\n");
+    fprintf(stderr, "   -T             enable Perl Tainting\n");
+    fprintf(stderr, "   -w             enable Perl Warnings\n");
     fprintf(stderr, "   -c             run syntax check only and exit (no execution)\n");
     fprintf(stderr, "   -r             display ePerl README file\n");
     fprintf(stderr, "   -l             display ePerl license files (COPYING and ARTISTIC)\n");
     fprintf(stderr, "   -v             display ePerl VERSION id\n");
     fprintf(stderr, "   -V             display ePerl VERSION id & compilation parameters\n");
-}
-
-char **set_variable(char **env, char *str) 
-{
-    char ca[1024];
-    char *cp;
-
-    strcpy(ca, str);
-    cp = strchr(ca, '=');
-    *cp++ = '\0';
-    return mysetenv(env, ca, cp);
-}
-
-char *remembered_perl_scalars[100] = { NULL };
-
-void remember_perl_scalar(char *str) 
-{
-    int i;
-
-    for (i = 0; remembered_perl_scalars[i] != NULL; i++)
-        ;
-    remembered_perl_scalars[i++] = strdup(str);
-    remembered_perl_scalars[i++] = NULL;
-    return;
-}
-
-void set_perl_scalars(void) 
-{
-    char ca[1024];
-    char *cp;
-    int i;
-
-    for (i = 0; remembered_perl_scalars[i] != NULL; i++) {
-        strcpy(ca, remembered_perl_scalars[i]);
-        cp = strchr(ca, '=');
-        *cp++ = '\0';
-        ePerl_SetScalar("main", ca, cp);
-    }
+    fprintf(stderr, "   -h             display ePerl usage list (this one)\n");
 }
 
 void mysighandler(int rc)
@@ -291,6 +271,7 @@ int main(int argc, char **argv, char **env)
     FILE *out = NULL;
     char *cpBuf = NULL;
     char *cpBuf2 = NULL;
+    char *cpBuf3 = NULL;
     char perlscript[1024] = "";
     char perlstderr[1024] = "";
     char perlstdout[1024] = "";
@@ -303,7 +284,7 @@ int main(int argc, char **argv, char **env)
     char *progname;
     int nBuf;
     int nOut;
-    char *source;
+    char *source = NULL;
     char sourcedir[2048];
     char *cp;
     static PerlInterpreter *my_perl = NULL; 
@@ -315,15 +296,22 @@ int main(int argc, char **argv, char **env)
     struct group *gr;
     int uid, gid;
     int keepcwd = FALSE;
-    char c;
+    int c;
     char *cpScript = NULL;
     int allow;
     int i, n, k;
     char *outputfile = NULL;
     char cwd[MAXPATHLEN];
     int fCheck = FALSE;
+    int fTaint = FALSE;
+    int fWarn = FALSE;
+    int fNoCase = FALSE;
+    int fPP = FALSE;
     char *cwd2;
     int fOkSwitch;
+    char *cpHost;
+    char *cpPort;
+    char *cpPath;
 
     /*  first step: our process initialisation */
     myinit();
@@ -334,14 +322,9 @@ int main(int argc, char **argv, char **env)
         progname = cp+1;
     }
 
-    /*  set behaviour for egetopt() function */
-    optneed  = (int)'+';    /* use '+' instead of ':' */
-    optmaybe = (int)'?';    /* use '?' for optional arguments */
-    optbad   = (int)'!';    /* use '!' for bad options */
-    optstart = "-+";
-
     /*  parse the option arguments */
-    while ((c = egetopt(argc, argv, "m+B+E+d+D+a+o+kxcvVrl")) != EOF) {
+    opterr = 0;
+    while ((c = getopt(argc, argv, ":m:B:E:id:D:a:o:kxI:PCLTwcvVrlh")) != -1) {
         if (optarg == NULL) 
             optarg = "(null)";
         switch (c) {
@@ -361,16 +344,19 @@ int main(int argc, char **argv, char **env)
                 }
                 break;
             case 'B':
-                begin_delimiter = strdup(optarg);
+                ePerl_begin_delimiter = strdup(optarg);
                 break;
             case 'E':
-                end_delimiter = strdup(optarg);
+                ePerl_end_delimiter = strdup(optarg);
+                break;
+            case 'i':
+                fNoCase = TRUE;
                 break;
             case 'd':
-                remember_perl_scalar(optarg);
+                Perl5_RememberScalar(optarg);
                 break;
             case 'D':
-                env = set_variable(env, optarg);
+                env = Perl5_SetEnvVar(env, optarg);
                 break;
             case 'o':
                 outputfile = strdup(optarg);
@@ -381,6 +367,24 @@ int main(int argc, char **argv, char **env)
             case 'x':
                 fDebug = TRUE;
                 break;
+            case 'I':
+                Perl5_RememberINC(optarg);
+                break;
+            case 'P':
+                fPP = TRUE;
+                break;
+            case 'C':
+                ePerl_convert_entities = TRUE;
+                break;
+            case 'L':
+                ePerl_line_continuation = TRUE;
+                break;
+            case 'T':
+                fTaint = TRUE;
+                break;
+            case 'w':
+                fWarn = TRUE;
+                break;
             case 'c':
                 fCheck = TRUE;
                 break;
@@ -396,7 +400,21 @@ int main(int argc, char **argv, char **env)
             case 'l':
                 give_license();
                 myexit(EX_OK);
-            case '!':
+            case 'h':
+                give_usage(progname);
+                myexit(EX_OK);
+            case '?':
+                if (isprint(optopt))
+                    fprintf(stderr, "%s: Unknown option `-%c'.\n\n", progname, optopt);
+                else
+                    fprintf(stderr, "%s: Unknown option character `\\x%x'.\n\n", progname, optopt);
+                give_usage(progname);
+                myexit(EX_USAGE);
+            case ':':
+                if (isprint(optopt))
+                    fprintf(stderr, "%s: Missing argument for option `-%c'.\n\n", progname, optopt);
+                else
+                    fprintf(stderr, "%s: Missing argument for option character `\\x%x'.\n\n", progname, optopt);
                 give_usage(progname);
                 myexit(EX_USAGE);
         }
@@ -448,33 +466,53 @@ int main(int argc, char **argv, char **env)
         else {
             mode = (mode == MODE_UNKNOWN ? MODE_CGI : mode);
         }
+
+        /* set the command line for ps output */
+        sprintf(ca, "%s %s [%sCGI]", argv[0], source, mode == MODE_NPHCGI ? "NPH-" : "");
+        argv[0] = strdup(ca);
     }
     else {
         /* else we are used in a wrong way... */
+        fprintf(stderr, "%s: Missing required file to process\n", progname);
+        fprintf(stderr, "%s: Use either a filename, ``-'' for stdin or PATH_TRANSLATED.\n\n", progname);
         give_usage(progname);
         myexit(EX_USAGE);
     }
 
     /* set default delimiters */
-    if (begin_delimiter == NULL) {
+    if (ePerl_begin_delimiter == NULL) {
         if (mode == MODE_FILTER)
-            begin_delimiter = BEGIN_DELIMITER_FILTER;
+            ePerl_begin_delimiter = BEGIN_DELIMITER_FILTER;
         else
-            begin_delimiter = BEGIN_DELIMITER_CGI;
+            ePerl_begin_delimiter = BEGIN_DELIMITER_CGI;
     }
-    if (end_delimiter == NULL) {
+    if (ePerl_end_delimiter == NULL) {
         if (mode == MODE_FILTER)
-            end_delimiter = END_DELIMITER_FILTER;
+            ePerl_end_delimiter = END_DELIMITER_FILTER;
         else
-            end_delimiter = END_DELIMITER_CGI;
+            ePerl_end_delimiter = END_DELIMITER_CGI;
     }
+    if (fNoCase)
+        ePerl_case_sensitive_delimiters = FALSE;
+    else
+        ePerl_case_sensitive_delimiters = TRUE;
 
-    /* the builtin GIF images */
+    /* the built-in GIF images */
     if ((mode == MODE_CGI || mode == MODE_NPHCGI) && (cp = getenv("PATH_INFO")) != NULL) { 
         if (strcmp(cp, "/logo.gif") == 0) {
             give_img_logo();
             myexit(0);
         }
+        else if (strcmp(cp, "/powered.gif") == 0) {
+            give_img_powered();
+            myexit(0);
+        }
+    }
+
+    /* CGI modes imply Preprocessor usage and HTML entity conversions */
+    if (mode == MODE_CGI || mode == MODE_NPHCGI) {
+        fPP = TRUE;
+        ePerl_convert_entities = TRUE;
     }
 
     /* check for valid source file */
@@ -509,6 +547,20 @@ int main(int argc, char **argv, char **env)
             }
         }
 
+        /*
+         *
+         *  == Perl Security ==
+         *
+         */
+
+        /* perhaps force Taint mode */
+        if (CGI_MODES_FORCE_TAINTING)
+            fTaint = TRUE;
+
+        /* perhaps force Warnings */
+        if (CGI_MODES_FORCE_WARNINGS)
+            fWarn = TRUE;
+
         /*
          *
          * == UID/GID switching ==
@@ -699,6 +751,27 @@ int main(int argc, char **argv, char **env)
         cpScript = cpBuf;
 
     /* now set the additional env vars */
+    env = mysetenv(env, "SCRIPT_SRC_PATH", "%s", abspath(source));
+    env = mysetenv(env, "SCRIPT_SRC_PATH_FILE", "%s", filename(source));
+    env = mysetenv(env, "SCRIPT_SRC_PATH_DIR", "%s", abspath(dirname(source)));
+    if ((cpPath = getenv("PATH_INFO")) != NULL) {
+        if ((cpHost = getenv("SERVER_NAME")) == NULL)
+            cpHost = "localhost";
+        cpPort = getenv("SERVER_PORT");
+        if (strcmp(cpPort, "80") == 0)
+            cpPort = NULL;
+        sprintf(ca, "http://%s%s%s%s", 
+                cpHost, cpPort != NULL ? ":" : "", cpPort != NULL ? cpPort : "", cpPath);
+        env = mysetenv(env, "SCRIPT_SRC_URL", "%s", ca);
+        env = mysetenv(env, "SCRIPT_SRC_URL_FILE", "%s", filename(ca));
+        env = mysetenv(env, "SCRIPT_SRC_URL_DIR", "%s", dirname(ca));
+    }
+    else {
+        env = mysetenv(env, "SCRIPT_SRC_URL", "file://%s", abspath(source));
+        env = mysetenv(env, "SCRIPT_SRC_URL_FILE", "%s", filename(source));
+        env = mysetenv(env, "SCRIPT_SRC_URL_DIR", "file://%s", abspath(source));
+    }
+
     env = mysetenv(env, "SCRIPT_SRC_SIZE", "%d", nBuf);
     stat(source, &st);
     env = mysetenv(env, "SCRIPT_SRC_MODIFIED", "%d", st.st_mtime);
@@ -709,10 +782,19 @@ int main(int argc, char **argv, char **env)
     pw = getpwuid(st.st_uid);
     env = mysetenv(env, "SCRIPT_SRC_OWNER", "%s", pw->pw_name);
     env = mysetenv(env, "VERSION_INTERPRETER", "%s", ePerl_WebID);
-    env = mysetenv(env, "VERSION_LANGUAGE", "Perl/%s", AC_perlvers);
+    env = mysetenv(env, "VERSION_LANGUAGE", "Perl/%s", AC_perl_vers);
+
+    /* optionally run the ePerl preprocessor */
+    if (fPP) {
+        if ((cpBuf3 = ePerl_PP(cpScript, Perl5_RememberedINC)) == NULL) {
+            PrintError(mode, source, NULL, NULL, "Preprocessing failed for `%s': %s", source, ePerl_PP_GetError());
+            CU(mode == MODE_FILTER ? EX_IOERR : EX_OK);
+        }
+        cpScript = cpBuf3;
+    }
 
     /* convert bristled source to valid Perl code */
-    if ((cpBuf2 = ePerl_Bristled2Perl(cpScript)) == NULL) {
+    if ((cpBuf2 = ePerl_Bristled2Plain(cpScript)) == NULL) {
         PrintError(mode, source, NULL, NULL, "Cannot convert bristled code file `%s' to pure HTML", source);
         CU(mode == MODE_FILTER ? EX_IOERR : EX_OK);
     }
@@ -781,9 +863,13 @@ int main(int argc, char **argv, char **env)
                only _parsed_, not evaluated/executed! */
     myargc = 0;
     myargv[myargc++] = progname;
+    if (fTaint) 
+        myargv[myargc++] = "-T";
+    if (fWarn) 
+        myargv[myargc++] = "-w";
     myargv[myargc++] = perlscript;   
 #ifdef HAVE_PERL_DYNALOADER
-    rc = perl_parse(my_perl, ePerl_xs_init, myargc, myargv, env);
+    rc = perl_parse(my_perl, Perl5_XSInit, myargc, myargv, env);
 #else
     rc = perl_parse(my_perl, NULL, myargc, myargv, env);
 #endif
@@ -803,6 +889,7 @@ int main(int argc, char **argv, char **env)
             CU(mode == MODE_FILTER ? EX_IOERR : EX_OK);
         }
     }
+
     /* Stop when we are just doing a syntax check */
     if (fCheck && mode == MODE_FILTER) {
         fclose(er);
@@ -828,9 +915,15 @@ int main(int argc, char **argv, char **env)
         chdir(sourcedir);
     }
 
+    /*  Set the previously remembered Perl 5 scalars (option -d) */
+    Perl5_SetRememberedScalars();
+    /*  Set the previously remembered Perl 5 INC entries (option -I) */
+    Perl5_SetRememberedINC();
+
+    /*  Force unbuffered I/O */
+    Perl5_ForceUnbufferedStdout();
+
     /*  NOW IT IS TIME to evaluate/execute the script!!! */
-    set_perl_scalars();
-    ePerl_ForceUnbufferedStdout();
     rc = perl_run(my_perl);
 
     /*  pre-close the handles, to be able to check
@@ -838,7 +931,6 @@ int main(int argc, char **argv, char **env)
     fclose(out); out = NULL;
     fclose(er);  er  = NULL;
 
-
     /*  when the Perl interpreter failed or there
         is data on stderr, we print a error page */
     if (stat(perlstderr, &st) == 0)
@@ -861,29 +953,29 @@ int main(int argc, char **argv, char **env)
     /*  if we are running as a NPH-CGI/1.1 script
         we had to provide the HTTP reponse headers ourself */
     if (mode == MODE_NPHCGI) {
-        PrintHTTPResponse();
+        HTTP_PrintResponseHeaders();
 
         /* if there are no HTTP header lines, we print a basic
            Content-Type header which should be ok */
-        if (!ePerl_HeadersExists(cpOut)) {
+        if (!HTTP_HeadersExists(cpOut)) {
             printf("Content-Type: text/html\n");
             printf("Content-Length: %d\n", nOut);
             printf("\n");
         }
     }
     else if (mode == MODE_CGI) {
-        ePerl_StripHTTPHeaders(&cpOut, &nOut);
+        HTTP_StripResponseHeaders(&cpOut, &nOut);
 
         /* if there are no HTTP header lines, we print a basic
            Content-Type header which should be ok */
-        if (!ePerl_HeadersExists(cpOut)) {
+        if (!HTTP_HeadersExists(cpOut)) {
             printf("Content-Type: text/html\n");
             printf("Content-Length: %d\n", nOut);
             printf("\n");
         }
     }
     else if (mode == MODE_FILTER) {
-        ePerl_StripHTTPHeaders(&cpOut, &nOut);
+        HTTP_StripResponseHeaders(&cpOut, &nOut);
     }
 
     /* ok, now recover the stdout and stderr and
@@ -940,6 +1032,16 @@ int main(int argc, char **argv, char **env)
     if (cpOut)
         free(cpOut);
 
+    /* remove temporary files */
+#ifndef DEBUG_ENABLED
+    if (*perlstderr != NUL)
+        unlink(perlstderr);
+    if (*perlstdout != NUL)
+        unlink(perlstdout);
+    if (*perlscript != NUL)
+        unlink(perlscript);
+#endif
+
     myexit(EXRC);
     return EXRC; /* make -Wall happy ;-) */
 }
